from . import ugettext


class DjangoOpenpayError(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class OpenpayObjectDoesNotExist(DjangoOpenpayError):
    def __init__(self):
        super(OpenpayObjectDoesNotExist, self).__init__(
            ugettext("This object's code does not exists inside the"
                     " OpenPay API.")
        )


class OpenpayNoCustomer(DjangoOpenpayError):
    def __init__(self):
        super(OpenpayNoCustomer, self).__init__(
            ugettext("This object does not have a related customer code and "
                     "cannot be saved in the OpenPay API.")
        )


class OpenpayNoCard(DjangoOpenpayError):
    def __init__(self):
        super(OpenpayNoCard, self).__init__(
            ugettext("This object does not have a related card code and "
                     "cannot be saved in the OpenPay API.")
        )


class OpenpayNotUserCard(DjangoOpenpayError):
    def __init__(self):
        super(OpenpayNotUserCard, self).__init__(
            ugettext("This Card does not belong to this Customer.")
        )



# Obtained and edited from:
# https://goo.gl/0lW3Di
from django.apps import apps as django_apps
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.http import HttpResponse

import base64
import binascii
from urllib.parse import unquote_plus

from . import ugettext_lazy


def get_customer_model():
    """
    Returns the Customer model that is active in this project.
    """
    try:
        return django_apps.get_model(settings.OPENPAY_CUSTOMER_MODEL)
    except ValueError:
        raise ImproperlyConfigured(
            "OPENPAY_CUSTOMER_MODEL must be of the form 'app_label.model_name'"
        )
    except LookupError:
        raise ImproperlyConfigured(
            "OPENPAY_CUSTOMER_MODEL refers to model '%s' "
            "that has not been installed" % settings.OPENPAY_CUSTOMER_MODEL
        )


class HttpResponseUnauthorized(HttpResponse):
    status_code = 401

    def __init__(self):
        super(HttpResponseUnauthorized, self).__init__()
        realm = getattr(settings, 'BASICAUTH_REALM', 'Secure resource')
        self['WWW-Authenticate'] = 'Basic realm="{}"'.format(realm)


def extract_basicauth(authorization_header, encoding='utf-8'):
    splitted = authorization_header.split(' ')
    if len(splitted) != 2:
        return None

    auth_type, auth_string = splitted

    if 'basic' != auth_type.lower():
        return None

    try:
        b64_decoded = base64.b64decode(auth_string)
    except (TypeError, binascii.Error):
        return None
    try:
        auth_string_decoded = b64_decoded.decode(encoding)
    except UnicodeDecodeError:
        return None

    splitted = auth_string_decoded.split(':')

    if len(splitted) != 2:
        return None

    username, password = map(unquote_plus, splitted)
    return username, password


def validate_request(request):
    # Check an incoming request.
    # Returns:
    #     - None if authentication failed
    #     - the validated username otherwise
    if not getattr(settings, 'OPENPAY_BASICAUTH_USERS', None):
        return True

    if 'HTTP_AUTHORIZATION' not in request.META:
        return None

    authorization_header = request.META['HTTP_AUTHORIZATION']
    ret = extract_basicauth(authorization_header)
    if not ret:
        return None

    username, password = ret

    if settings.OPENPAY_BASICAUTH_USERS.get(username) != password:
        return None

    return username



from django.utils.dateparse import parse_datetime, parse_date
import logging

from . import models

logger = logging.getLogger(__name__)


def verification(body):
    logger.info(body['verification_code'])


def chargeRefunded(body):
    transaction = body['transaction']
    card = None
    customer = None
    subscription = None
    if 'customer_id' in transaction:
        customer = models.get_customer_model().objects.get(
            openpay_id=transaction['customer_id'])
    if 'subscription_id' in transaction:
        subscription = models.Subscription.objects.get(
            openpay_id=transaction['subscription_id'])
    if 'card' in transaction:
        card = models.Card.objects.get(
            openpay_id=transaction['card']['id'])
        if not customer and 'customer_id' in transaction['card']:
            customer = models.get_customer_model().objects.get(
                openpay_id=transaction['card']['customer_id'])
    charge, created = models.Charge.objects.update_or_create(
        openpay_id=transaction['id'],
        defaults={
            'authorization': transaction['authorization'],
            'method': transaction['method'],
            'operation_type': transaction['operation_type'],
            'transaction_type': transaction['transaction_type'],
            'status': transaction['status'],
            'conciliated': transaction['conciliated'],
            'creation_date': parse_datetime(transaction['creation_date']),
            'operation_date': parse_datetime(transaction['operation_date']),
            'description': transaction['description'],
            'error_message': transaction['error_message'],
            'order_id': transaction['order_id'],
            'amount': transaction['amount'],
            'currency': transaction['currency'],
            'customer': customer,
            'card': card,
        })

    transaction = transaction['refund']
    refund, created = models.Refund.objects.update_or_create(
        openpay_id=transaction['id'],
        defaults={
            'authorization': transaction['authorization'],
            'method': transaction['method'],
            'operation_type': transaction['operation_type'],
            'transaction_type': transaction['transaction_type'],
            'status': transaction['status'],
            'conciliated': transaction['conciliated'],
            'creation_date': parse_datetime(transaction['creation_date']),
            'operation_date': parse_datetime(transaction['operation_date']),
            'description': transaction['description'],
            'error_message': transaction['error_message'],
            'order_id': transaction['order_id'],
            'amount': transaction['amount'],
            'currency': transaction['currency'],
            'customer': customer,
            'charge': charge,
        })


def chargeCancelled(body):
    transaction = body['transaction']
    card = None
    customer = None
    subscription = None
    if 'customer_id' in transaction:
        customer = models.get_customer_model().objects.get(
            openpay_id=transaction['customer_id'])
    if 'subscription_id' in transaction:
        subscription = models.Subscription.objects.get(
            openpay_id=transaction['subscription_id'])
    if 'card' in transaction:
        card = models.Card.objects.get(
            openpay_id=transaction['card']['id'])
        if not customer and 'customer_id' in transaction['card']:
            customer = models.get_customer_model().objects.get(
                openpay_id=transaction['card']['customer_id'])
    charge, created = models.Charge.objects.update_or_create(
        openpay_id=transaction['id'],
        defaults={
            'authorization': transaction['authorization'],
            'method': transaction['method'],
            'operation_type': transaction['operation_type'],
            'transaction_type': transaction['transaction_type'],
            'status': transaction['status'],
            'conciliated': transaction['conciliated'],
            'creation_date': parse_datetime(transaction['creation_date']),
            'operation_date': parse_datetime(transaction['operation_date']),
            'description': transaction['description'],
            'error_message': transaction['error_message'],
            'order_id': transaction['order_id'],
            'amount': transaction['amount'],
            'currency': transaction['currency'],
            'customer': customer,
            'card': card,
            'subscription': subscription
        })


def chargeCreated(body):
    transaction = body['transaction']
    card = None
    customer = None
    subscription = None
    if 'customer_id' in transaction:
        customer = models.get_customer_model().objects.get(
            openpay_id=transaction['customer_id'])
    if 'subscription_id' in transaction:
        subscription = models.Subscription.objects.get(
            openpay_id=transaction['subscription_id'])
    if 'card' in transaction:
        card = models.Card.objects.get(
            openpay_id=transaction['card']['id'])
        if not customer and 'customer_id' in transaction['card']:
            customer = models.get_customer_model().objects.get(
                openpay_id=transaction['card']['customer_id'])
    charge, created = models.Charge.objects.update_or_create(
        openpay_id=transaction['id'],
        defaults={
            'authorization': transaction['authorization'],
            'method': transaction['method'],
            'operation_type': transaction['operation_type'],
            'transaction_type': transaction['transaction_type'],
            'status': transaction['status'],
            'conciliated': transaction['conciliated'],
            'creation_date': parse_datetime(transaction['creation_date']),
            'operation_date': parse_datetime(transaction['operation_date']),
            'description': transaction['description'],
            'error_message': transaction['error_message'],
            'order_id': transaction['order_id'],
            'amount': transaction['amount'],
            'currency': transaction['currency'],
            'customer': customer,
            'card': card,
            'subscription': subscription
        })


def chargeSucceeded(body):
    transaction = body['transaction']
    card = None
    customer = None
    subscription = None
    if 'customer_id' in transaction:
        customer = models.get_customer_model().objects.get(
            openpay_id=transaction['customer_id'])
    if 'subscription_id' in transaction:
        subscription = models.Subscription.objects.get(
            openpay_id=transaction['subscription_id'])
    if 'card' in transaction:
        card = models.Card.objects.get(
            openpay_id=transaction['card']['id'])
        if not customer and 'customer_id' in transaction['card']:
            customer = models.get_customer_model().objects.get(
                openpay_id=transaction['card']['customer_id'])
    charge, created = models.Charge.objects.update_or_create(
        openpay_id=transaction['id'],
        defaults={
            'authorization': transaction['authorization'],
            'method': transaction['method'],
            'operation_type': transaction['operation_type'],
            'transaction_type': transaction['transaction_type'],
            'status': transaction['status'],
            'conciliated': transaction['conciliated'],
            'creation_date': parse_datetime(transaction['creation_date']),
            'operation_date': parse_datetime(transaction['operation_date']),
            'description': transaction['description'],
            'error_message': transaction['error_message'],
            'order_id': transaction['order_id'],
            'amount': transaction['amount'],
            'currency': transaction['currency'],
            'customer': customer,
            'card': card,
            'subscription': subscription
        })



from django.db.models import Q

from celery.decorators import periodic_task
from celery.task.schedules import crontab
from celery.utils.log import get_task_logger
from datetime import datetime

from . import models, hardcode

logger = get_task_logger(__name__)


@periodic_task(
    run_every=(crontab(hour=23, minute=59)),
    name="updateSubscriptions",
    ignore_result=True
)
def updateSubscriptions():
    logger.info("Updating Subscriptions information from Openpay's servers...")

    subscriptions_in_system = models.Subscription.objects.filter(
        Q(charge_date__gte=datetime.now().date()) | Q(charge_date=None)
    ).exclude(
        status=hardcode.subscription_status_cancelled)

    if subscriptions_in_system.exists():
        for subscription in subscriptions_in_system.iterator():
            subscription.op_refresh(save=True)

    logger.info("System subscriptions up to date with Openpay's servers.")



from django.contrib import admin
from . import models, ugettext
from .utils import get_customer_model

CustomerModel = get_customer_model()


# Register your models here.
@admin.register(models.Address)
class AddressAdmin(admin.ModelAdmin):
    model = models.Address
    list_display = ('line1', 'line2', 'line3', 'city', 'state',
                    'country_code', 'postal_code')

    def get_readonly_fields(self, request, obj=None):
        return models.Address.get_readonly_fields(obj)


@admin.register(models.Card)
class CardAdmin(admin.ModelAdmin):
    model = models.Card
    actions = ['refresh', 'dismiss', ]
    list_display = ('pk', 'openpay_id', 'alias', 'holder', 'customer',
                    'creation_date', 'deleted')

    def refresh(self, request, queryset):
        refreshed = 0
        for instance in queryset:
            instance.skip_signal = True
            instance.op_refresh(save=True)
            refreshed = refreshed + 1
        if captured == 1:
            message_bit = "1 instance was"
        else:
            message_bit = "%s instances were" % refreshed
        self.message_user(
            request,
            "%s successfully refreshed." % message_bit
        )
    refresh.short_description = ugettext('Refresh selected instances')

    def dismiss(self, request, queryset):
        dismissed = 0
        for instance in queryset:
            instance.skip_signal = True
            instance.op_dismiss(save=True)
            dismissed = dismissed + 1
        if dismissed == 1:
            message_bit = "1 instance was"
        else:
            message_bit = "%s instances were" % dismissed
        self.message_user(
            request,
            "%s successfully dismissed." % message_bit
        )
    dismiss.short_description = ugettext('Dismiss selected instances')

    def get_readonly_fields(self, request, obj=None):
        return models.Card.get_readonly_fields(obj)


@admin.register(models.Plan)
class PlanAdmin(admin.ModelAdmin):
    model = models.Plan
    actions = ['refresh', 'dismiss', ]
    list_display = ('name', 'openpay_id', 'status', 'amount', 'repeat_every',
                    'repeat_unit', 'creation_date', 'deleted')

    def refresh(self, request, queryset):
        refreshed = 0
        for instance in queryset:
            instance.skip_signal = True
            instance.op_refresh(save=True)
            refreshed = refreshed + 1
        if captured == 1:
            message_bit = "1 instance was"
        else:
            message_bit = "%s instances were" % refreshed
        self.message_user(
            request,
            "%s successfully refreshed." % message_bit
        )
    refresh.short_description = ugettext('Refresh selected instances')

    def dismiss(self, request, queryset):
        dismissed = 0
        for instance in queryset:
            instance.skip_signal = True
            instance.op_dismiss(save=True)
            dismissed = dismissed + 1
        if dismissed == 1:
            message_bit = "1 instance was"
        else:
            message_bit = "%s instances were" % dismissed
        self.message_user(
            request,
            "%s successfully dismissed." % message_bit
        )
    dismiss.short_description = ugettext('Dismiss selected instances')

    def get_readonly_fields(self, request, obj=None):
        return models.Plan.get_readonly_fields(obj)


@admin.register(models.Subscription)
class SubscriptionAdmin(admin.ModelAdmin):
    model = models.Subscription
    actions = ['refresh', 'dismiss', ]
    list_display = ('pk', 'openpay_id', 'customer', 'plan', 'card', 'status',
                    'creation_date', 'deleted')

    def refresh(self, request, queryset):
        refreshed = 0
        for instance in queryset:
            instance.skip_signal = True
            instance.op_refresh(save=True)
            refreshed = refreshed + 1
        if refreshed == 1:
            message_bit = "1 instance was"
        else:
            message_bit = "%s instances were" % refreshed
        self.message_user(
            request,
            "%s successfully refreshed." % message_bit
        )
    refresh.short_description = ugettext('Refresh selected instances')

    def dismiss(self, request, queryset):
        dismissed = 0
        for instance in queryset:
            instance.skip_signal = True
            instance.op_dismiss(save=True)
            dismissed = dismissed + 1
        if dismissed == 1:
            message_bit = "1 instance was"
        else:
            message_bit = "%s instances were" % dismissed
        self.message_user(
            request,
            "%s successfully dismissed." % message_bit
        )
    dismiss.short_description = ugettext('Dismiss selected instances')

    def get_readonly_fields(self, request, obj=None):
        return models.Subscription.get_readonly_fields(obj)


@admin.register(models.Refund)
class RefundAdmin(admin.ModelAdmin):
    model = models.Refund
    actions = ['refresh', 'dismiss', ]
    list_display = ('pk', 'openpay_id', 'customer', 'charge', 'amount',
                    'creation_date', 'deleted')

    def refresh(self, request, queryset):
        refreshed = 0
        for instance in queryset:
            instance.skip_signal = True
            instance.op_refresh(save=True)
            refreshed = refreshed + 1
        if captured == 1:
            message_bit = "1 instance was"
        else:
            message_bit = "%s instances were" % refreshed
        self.message_user(
            request,
            "%s successfully refreshed." % message_bit
        )
    refresh.short_description = ugettext('Refresh selected instances')

    def dismiss(self, request, queryset):
        dismissed = 0
        for instance in queryset:
            instance.skip_signal = True
            instance.op_dismiss(save=True)
            dismissed = dismissed + 1
        if dismissed == 1:
            message_bit = "1 instance was"
        else:
            message_bit = "%s instances were" % dismissed
        self.message_user(
            request,
            "%s successfully dismissed." % message_bit
        )
    dismiss.short_description = ugettext('Dismiss selected instances')

    def get_readonly_fields(self, request, obj=None):
        return models.Refund.get_readonly_fields(obj)


@admin.register(models.Charge)
class ChargeAdmin(admin.ModelAdmin):
    model = models.Charge
    actions = ['refresh', 'capture', 'refund', 'dismiss', ]
    list_display = ('pk', 'openpay_id', 'customer', 'card', 'amount',
                    'creation_date', 'deleted')

    def refresh(self, request, queryset):
        refreshed = 0
        for instance in queryset:
            instance.skip_signal = True
            instance.op_refresh(save=True)
            refreshed = refreshed + 1
        if captured == 1:
            message_bit = "1 instance was"
        else:
            message_bit = "%s instances were" % refreshed
        self.message_user(
            request,
            "%s successfully refreshed." % message_bit
        )
    refresh.short_description = ugettext('Refresh selected instances')

    def capture(self, request, queryset):
        captured = 0
        for charge in queryset:
            charge.op_capture()
            captured = captured + 1
        if captured == 1:
            message_bit = "1 charge was"
        else:
            message_bit = "%s charges were" % captured
        self.message_user(
            request,
            "%s successfully captured." % message_bit
        )
    capture.short_description = ugettext('Capture selected charges')

    def refund(self, request, queryset):
        refunded = 0
        for charge in queryset:
            charge.op_refund()
            refunded = refunded + 1
        if refunded == 1:
            message_bit = "1 charge was"
        else:
            message_bit = "%s charges were" % refunded
            self.message_user(
                request,
                "%s successfully refunded." % message_bit
            )
    refund.short_description = ugettext('Refund selected charges')

    def dismiss(self, request, queryset):
        dismissed = 0
        for instance in queryset:
            instance.skip_signal = True
            instance.op_dismiss(save=True)
            dismissed = dismissed + 1
        if dismissed == 1:
            message_bit = "1 instance was"
        else:
            message_bit = "%s instances were" % dismissed
        self.message_user(
            request,
            "%s successfully dismissed." % message_bit
        )
    dismiss.short_description = ugettext('Dismiss selected instances')

    def get_readonly_fields(self, request, obj=None):
        return models.Charge.get_readonly_fields(obj)



from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.translation import ugettext, ugettext_lazy, ungettext_lazy

import openpay

default_app_config = 'django_openpay.apps.DjangoOpenpayConfig'


def start():
    OPENPAY_PRIVATE_API_KEY = getattr(
        settings, 'OPENPAY_PRIVATE_API_KEY', None)
    OPENPAY_VERIFY_SSL = getattr(settings, 'OPENPAY_VERIFY_SSL', None)
    OPENPAY_MERCHANT_ID = getattr(settings, 'OPENPAY_MERCHANT_ID', None)
    OPENPAY_DEVICE_ID = getattr(settings, 'OPENPAY_DEVICE_ID', None)
    OPENPAY_PRODUCTION = getattr(settings, 'OPENPAY_PRODUCTION', None)
    # OPENPAY_BASICAUTH_USERS = getattr(
    #     settings, 'OPENPAY_BASICAUTH_USERS', None)
    OPENPAY_CUSTOMER_MODEL = getattr(
        settings, 'OPENPAY_CUSTOMER_MODEL', None)

    if not OPENPAY_PRIVATE_API_KEY:
        raise ImproperlyConfigured(
            "OPENPAY_PRIVATE_API_KEY must be defined. (String)"
        )
    if OPENPAY_VERIFY_SSL not in [True, False]:
        raise ImproperlyConfigured(
            "OPENPAY_VERIFY_SSL must be defined. (Boolean)"
        )
    if not OPENPAY_MERCHANT_ID:
        raise ImproperlyConfigured(
            "OPENPAY_MERCHANT_ID must be defined. (String)"
        )
    if not OPENPAY_DEVICE_ID or not (0 <= len(OPENPAY_DEVICE_ID) <= 31):
        raise ImproperlyConfigured(
            "OPENPAY_DEVICE_ID must be defined and must have a length between "
            "1 and 32 characters. (String)"
        )
    if OPENPAY_PRODUCTION not in [True, False]:
        raise ImproperlyConfigured(
            "OPENPAY_PRODUCTION must be defined. (Boolean)"
        )
    # if not OPENPAY_BASICAUTH_USERS:
    #     raise ImproperlyConfigured(
    #         "OPENPAY_BASICAUTH_USERS must be defined. (Dict)"
    #     )
    if not OPENPAY_CUSTOMER_MODEL:
        raise ImproperlyConfigured(
            "OPENPAY_CUSTOMER_MODEL must be defined. (String)"
        )

    openpay.api_key = OPENPAY_PRIVATE_API_KEY
    openpay.verify_ssl_certs = OPENPAY_VERIFY_SSL
    openpay.merchant_id = OPENPAY_MERCHANT_ID
    openpay.device_id = OPENPAY_DEVICE_ID
    openpay.production = OPENPAY_PRODUCTION

start()



from django.db import transaction
from django.core.management.base import BaseCommand
from django.conf import settings
from django.utils.dateparse import parse_datetime, parse_date

from decimal import Decimal
import os
import random

import django_openpay as djop


class Command(BaseCommand):
    def plans(self):
        dbPlans = dict()
        print('Pulling Plans from Openpay ...')
        plansList = djop.openpay.Plan.all()
        if plansList.get('count', 0) > 0:
            for planJson in plansList.get('data', []):
                dbobj = None
                try:
                    dbobj = djop.models.Plan.objects.get(
                        openpay_id=planJson['id'])
                except djop.models.Plan.DoesNotExist:
                    dbobj = djop.models.Plan(
                        openpay_id=planJson['id'])
                finally:
                    dbobj.skip_signal = True
                    dbobj.name = planJson['name']
                    dbobj.amount = Decimal(planJson['amount'])
                    dbobj.currency = planJson['currency']
                    dbobj.status_after_retry = planJson['status_after_retry']
                    dbobj.retry_times = planJson['retry_times']
                    dbobj.repeat_unit = planJson['repeat_unit']
                    dbobj.trial_days = planJson['trial_days']
                    dbobj.repeat_every = planJson['repeat_every']
                    dbobj.status = planJson['status']
                    dbobj.creation_date = parse_datetime(
                        planJson['creation_date'])
                    dbobj.save()
                    dbPlans[planJson['id']] = dbobj.pk
        return dbPlans

    def cards(self, customer):
        dbCards = dict()
        cardsList = customer.op_cards()
        if cardsList.get('count', 0) > 0:
            for cardJson in cardsList.get('data', []):
                try:
                    dbobj = djop.models.Card.objects.get(
                        openpay_id=cardJson['id'])
                except djop.models.Card.DoesNotExist:
                    dbobj = djop.models.Card(
                        openpay_id=cardJson['id'])
                finally:
                    dbobj.skip_signal = True
                    dbobj.customer_id = customer.pk
                    dbobj.card_type = cardJson['type']
                    dbobj.holder = cardJson['holder_name']
                    dbobj.number = cardJson['card_number'][-4:]
                    dbobj.bank_name = cardJson['bank_name']
                    dbobj.brand = cardJson['brand']
                    dbobj.month = cardJson['expiration_month'][-2:]
                    dbobj.year = cardJson['expiration_year'][-2:]
                    dbobj.creation_date = parse_datetime(
                        cardJson['creation_date'])
                    dbobj.save()
                    dbCards[cardJson['id']] = dbobj.pk
        return dbCards

    def subscriptions(self, customer, plans, cards):
        subscriptionsList = customer.op_subscriptions()
        if subscriptionsList.get('count', 0) > 0:
            for subscriptionJson in subscriptionsList.get('data', []):
                try:
                    dbobj = djop.models.Subscription.objects.get(
                        openpay_id=subscriptionJson['id'])
                except djop.models.Subscription.DoesNotExist:
                    dbobj = djop.models.Subscription(
                        openpay_id=subscriptionJson['id'])
                finally:
                    dbobj.skip_signal = True
                    dbobj.customer_id = customer.pk
                    dbobj.plan_id = plans[subscriptionJson['plan_id']]
                    dbobj.card_id = cards[subscriptionJson['card']['id']]
                    dbobj.cancel_at_period_end = \
                        subscriptionJson['cancel_at_period_end']
                    new_charge_date = parse_date(
                        subscriptionJson['charge_date'])
                    dbobj.latest_charge_date = dbobj.charge_date if \
                        dbobj.charge_date != new_charge_date else \
                        dbobj.latest_charge_date
                    dbobj.charge_date = new_charge_date
                    dbobj.period_end_date = parse_date(
                        subscriptionJson['period_end_date'])
                    dbobj.status = subscriptionJson['status']
                    dbobj.current_period_number = \
                        subscriptionJson['current_period_number']
                    dbobj.trial_end_date = parse_date(
                        subscriptionJson['trial_end_date'])
                    dbobj.creation_date = parse_datetime(
                        subscriptionJson['creation_date'])
                    dbobj.save()
        return subscriptionsList.get('count', 0)

    def customers(self, plans):
        cardsNum = 0
        subsNum = 0
        print('Pulling Customers from Openpay ...')
        customerModel = djop.utils.get_customer_model()
        customersList = djop.openpay.Customer.all()
        if customersList.get('count', 0) > 0:
            for customerJson in customersList.get('data', []):
                try:
                    dbobj = customerModel.objects.get(
                        openpay_id=customerJson['id'])
                except customerModel.DoesNotExist:
                    dbobj = customerModel(openpay_id=customerJson['id'])
                finally:
                    dbobj.skip_signal = True
                    dbobj.op_refresh(save=True)
                    cards = self.cards(dbobj)
                    cardsNum += len(cards)
                    subs = self.subscriptions(dbobj, plans, cards)
                    subsNum += subs

        return customersList.get('count', 0), cardsNum, subsNum

    @transaction.atomic
    def handle(self, *args, **options):
        # print('Pulling Subscriptions from Openpay ...')
        # plansList = djop.openpay.Subscription.all()
        plans = self.plans()
        customers, cards, subs = self.customers(plans)

        print('{} plans pulled from Openpay.'.format(len(plans)))
        print('{} customers pulled from Openpay.'.format(customers))
        print('{} cards pulled from Openpay.'.format(cards))
        print('{} subscriptions pulled from Openpay.'.format(subs))



from django.conf.urls import url

from . import views

urlpatterns = [
    url(r'^save/$', views.cardSave,
        name='django_openpay_cardsave'),
    url(r'^webhook/$', views.webhook,
        name='django_openpay_webhook'),
]



# Obtained and edited from:
# https://goo.gl/0lW3Di

from functools import wraps

from .utils import validate_request, HttpResponseUnauthorized


def basic_auth_required(func):
    @wraps(func)
    def _wrapped(request, *args, **kwargs):
        validated_username = validate_request(request)
        if not validated_username:
            return HttpResponseUnauthorized()
        else:
            request.META['REMOTE_USER'] = validated_username
            return func(request, *args, **kwargs)
    return _wrapped


def skippable(signal_func):
    @wraps(signal_func)
    def _decorator(sender, instance, **kwargs):
        if getattr(instance, 'skip_signal', False):
            return None
        return signal_func(sender, instance, **kwargs)
    return _decorator



from django.conf import settings
from django.core.validators import RegexValidator
from django.db import models
from django.db.models.fields.related import ManyToManyField
from django.db.models.signals import pre_save, pre_delete, post_delete
from django.dispatch import receiver
from django.utils.dateparse import parse_datetime, parse_date

from decimal import Decimal
from jsonfield import JSONField

from . import openpay, hardcode, ugettext_lazy, exceptions, ungettext_lazy
from .decorators import skippable
from .utils import get_customer_model


phone_validator = RegexValidator(
    regex=r'^\d{9,15}$',
    message=ugettext_lazy("The telephone number can only contain digits. "
                          " The maximum number of digits is 15.")
)

CustomerModel = settings.OPENPAY_CUSTOMER_MODEL


class AbstractOpenpayBase(models.Model):
    openpay_id = models.CharField(
        max_length=100,
        blank=True,
        verbose_name=ugettext_lazy('Openpay ID'))
    # Not using auto_now_add because this is not the date from Django, but
    # the one from Openpay
    deleted = models.BooleanField(
        default=False,
        blank=True,
        null=False,
        verbose_name=ugettext_lazy('Deleted from Openpay'))
    creation_date = models.DateTimeField(
        blank=True,
        null=False,
        verbose_name=ugettext_lazy('Creation date'))

    class Meta:
        abstract = True

    @classmethod
    def get_readonly_fields(self, instance=None):
        raise NotImplementedError

    def op_commit(self):
        # Save the changes in the object directly to the openpay servers
        raise NotImplementedError

    def op_refresh(self, save=False):
        # Call the op_load, and op_fill always. This function is designed to
        # maintain the object updated with what is in the openpay server.
        self.op_load()
        if self._op_:
            self.op_fill()
        if save:
            self.save()

    def op_load(self):
        # Pull the Openpay data, always
        raise NotImplementedError

    def op_nones2str(self, items):
        for k in items:
            v = getattr(self._op_, k)
            if v is None:
                v = 'Desconocido'
                setattr(self._op_, k, v)

    def op_fill(self):
        # Only update the object's fields with the openpay data.
        # If the Openpay data has not been loaded, call op_load.
        raise NotImplementedError

    def op_dismiss(self, save=False):
        # Execute the removal of the openpay object (in the openpay server),
        # this same removal could be a logical or physical destruction, but the
        # way to call it is the same.
        if self.openpay_id:
            if not hasattr(self, '_op_'):
                self.op_load()

            if self._op_:
                self.op_fill()
                try:
                    self._op_.delete()
                    self.deleted = True
                except openpay.error.InvalidRequestError as e:
                    if e.http_status == 404:
                        self.deleted = True
                        self._op_ = None
                    else:
                        raise e

        if save:
            self.skip_signal = True
            self.save()


class Address(models.Model):
    city = models.TextField(
        blank=False,
        verbose_name=ugettext_lazy('City'))
    state = models.TextField(
        blank=False,
        verbose_name=ugettext_lazy('State'))
    line1 = models.CharField(
        max_length=100,
        blank=False,
        verbose_name=ugettext_lazy('Street (Line 1)'))
    line2 = models.CharField(
        max_length=100,
        blank=True,
        verbose_name=ugettext_lazy('Street (Line 2)'))
    line3 = models.CharField(
        max_length=100,
        blank=True,
        verbose_name=ugettext_lazy('Street (Line 3)'))
    postal_code = models.IntegerField(
        blank=False,
        null=False,
        verbose_name=ugettext_lazy('Postal Code'))
    country_code = models.CharField(
        choices=hardcode.address_countrycodes,
        default='MX',
        max_length=5,
        blank=True,
        verbose_name=ugettext_lazy('Country'))
    creation_date = models.DateTimeField(
        auto_now_add=True,
        verbose_name=ugettext_lazy('Creation date'))

    @classmethod
    def get_readonly_fields(self, instance=None):
        if instance:
            return ['creation_date']
        return ['creation_date']

    # Obtained and edited from:
    # https://goo.gl/SqkLbo
    @property
    def json_dict(self):
        opts = self._meta
        data = {}
        for f in opts.concrete_fields + opts.many_to_many:
            if isinstance(f, ManyToManyField):
                if self.pk is None:
                    data[f.name] = []
                else:
                    data[f.name] = list(
                        f.value_from_object(self).values_list('pk', flat=True)
                    )
            elif f.name not in ['id', 'creation_date']:
                data[f.name] = f.value_from_object(self)
        return data


# class Customer(AbstractOpenpayBase):
class AbstractCustomer(AbstractOpenpayBase):
    first_name = models.CharField(
        max_length=60,
        blank=False,
        verbose_name=ugettext_lazy('First Name'))
    last_name = models.CharField(
        max_length=60,
        blank=True,
        verbose_name=ugettext_lazy('Last Name'))
    email = models.EmailField(
        blank=False,
        null=False,
        verbose_name=ugettext_lazy('Email'))
    phone_number = models.CharField(
        validators=[phone_validator],
        max_length=15,
        blank=True,
        verbose_name=ugettext_lazy('Phone Number'))
    address = models.OneToOneField(
        Address,
        blank=True,
        null=True,
        related_name='customer',
        verbose_name=ugettext_lazy('Address'))

    class Meta:
        abstract = True

    @classmethod
    def get_readonly_fields(self, instance=None):
        if instance:
            return ['openpay_id', 'creation_date']
        return ['openpay_id', 'creation_date']

    def op_cards(self):
        if self.openpay_id:
            if not hasattr(self, '_op_'):
                self.op_load()
            if self._op_:
                return self._op_.cards.all()
        else:
            raise exceptions.OpenpayObjectDoesNotExist

    def op_subscriptions(self):
        if self.openpay_id:
            if not hasattr(self, '_op_'):
                self.op_load()
            if self._op_:
                return self._op_.subscriptions.all()
        else:
            raise exceptions.OpenpayObjectDoesNotExist

    def op_commit(self):
        if self.openpay_id:
            if not hasattr(self, '_op_'):
                self.op_load()
            if not self.deleted and self._op_:
                self._op_.name = self.first_name
                self._op_.last_name = self.last_name if \
                    self.last_name else None
                self._op_.email = self.email
                self._op_.phone_number = self.phone_number if \
                    self.phone_number else None
                self._op_.address = self.address.json_dict if \
                    self.address else None
                self._op_.save()

        else:
            self._op_ = openpay.Customer.create(
                name=self.first_name,
                last_name=self.last_name if self.last_name else None,
                email=self.email,
                phone_number=self.phone_number if self.phone_number else None,
                address=self.address.json_dict if self.address else None)
            self.openpay_id = self._op_.id
        self.op_fill()
        # We have a pre_save signal, so we dont need to save it manually

    def op_load(self):
        if self.openpay_id:
            try:
                self._op_ = openpay.Customer.retrieve(self.openpay_id)
            except openpay.error.InvalidRequestError as e:
                if e.http_status == 404:
                    self.deleted = True
                    self._op_ = None
                else:
                    raise e

        else:
            raise exceptions.OpenpayObjectDoesNotExist

    def op_fill(self):
        if not hasattr(self, '_op_'):
            self.op_load()
        if self._op_:
            self.op_nones2str(['name', 'last_name', 'phone_number'])
            self.first_name = self._op_.name
            self.last_name = self._op_.last_name
            self.email = self._op_.email
            self.phone_number = self._op_.phone_number
            self.creation_date = parse_datetime(
                self._op_.creation_date)

    @property
    def full_name(self):
        return '{first_name} {last_name}'.format(
            first_name=self.first_name,
            last_name=self.last_name)

    def __str__(self):
        return self.full_name


@receiver(pre_save, sender=CustomerModel)
@skippable
def customer_presave(sender, instance=None, **kwargs):
    instance.full_clean()
    instance.email = instance.email.strip()
    instance.op_commit()


@receiver(pre_delete, sender=CustomerModel)
@skippable
def customer_postdelete(sender, instance, **kwargs):
    instance.op_dismiss()


class Card(AbstractOpenpayBase):
    alias = models.CharField(
        max_length=100,
        blank=True,
        verbose_name=ugettext_lazy('Alias'))
    card_type = models.CharField(
        max_length=15,
        blank=False,
        verbose_name=ugettext_lazy('Card type'))
    holder = models.CharField(
        max_length=100,
        blank=True,
        verbose_name=ugettext_lazy('Holder name'))
    number = models.CharField(
        max_length=5,
        blank=False,
        verbose_name=ugettext_lazy('Number'))
    bank_name = models.CharField(
        default='',
        max_length=30,
        blank=False,
        verbose_name=ugettext_lazy('Bank name'))
    brand = models.CharField(
        default=hardcode.card_brands_unknown,
        choices=hardcode.card_brands,
        max_length=20,
        blank=False,
        verbose_name=ugettext_lazy('Brand'))
    month = models.CharField(
        max_length=3,
        blank=True,
        verbose_name=ugettext_lazy('Expiration month'))
    year = models.CharField(
        max_length=3,
        blank=True,
        verbose_name=ugettext_lazy('Expiration year'))
    customer = models.ForeignKey(
        CustomerModel,
        blank=False,
        related_name='cards',
        verbose_name=ugettext_lazy('Owner'))

    @classmethod
    def get_readonly_fields(self, instance=None):
        if instance:
            return ['openpay_id', 'card_type', 'holder', 'number', 'month',
                    'bank_name', 'brand', 'year', 'customer', 'creation_date']
        return ['openpay_id', 'card_type', 'holder', 'number', 'month', 'year',
                'bank_name', 'brand', 'customer', 'creation_date']

    @classmethod
    def create_with_token(cls, customerId, tokenId, deviceId, alias=''):
        card_op = openpay.Card.create(
            customer=customerId, token_id=tokenId, device_session_id=deviceId)
        customer = get_customer_model().objects.get(openpay_id=customerId)
        # The card addres cannot be consulted
        card = cls(
            openpay_id=card_op.id,
            alias=alias,
            customer=customer)
        card._op_ = card_op
        card.op_fill()
        card.save()
        return card

    def op_commit(self):
        raise NotImplementedError

    def op_load(self):
        if not self.customer or not self.customer.openpay_id:
            raise exceptions.OpenpayNoCustomer

        if self.openpay_id:
            try:
                self._op_ = openpay.Customer.retrieve(
                    self.customer.openpay_id
                ).cards.retrieve(
                    self.openpay_id
                )
            except openpay.error.InvalidRequestError as e:
                if e.http_status == 404:
                    self.deleted = True
                    self._op_ = None
                else:
                    raise e

        else:
            raise exceptions.OpenpayObjectDoesNotExist

    def op_fill(self):
        if not hasattr(self, '_op_'):
            self.op_load()
        if self._op_:
            self.op_nones2str([
                'type', 'holder_name', 'bank_name', 'expiration_month',
                'expiration_year'])
            self.card_type = self._op_.type
            self.holder = self._op_.holder_name
            self.number = self._op_.card_number[-4:]
            self.bank_name = self._op_.bank_name
            self.brand = self._op_.brand
            self.month = self._op_.expiration_month[-2:]
            self.year = self._op_.expiration_year[-2:]
            self.creation_date = parse_datetime(
                self._op_.creation_date)

    def __str__(self):
        if self.alias:
            return self.alias
        return '{customer}-{pk}'.format(customer=self.customer, pk=self.pk)


@receiver(pre_save, sender=Card)
@skippable
def card_presave(sender, instance=None, **kwargs):
    instance.full_clean()


@receiver(post_delete, sender=Card)
@skippable
def card_postdelete(sender, instance, **kwargs):
    instance.op_dismiss()


class Plan(AbstractOpenpayBase):
    name = models.CharField(
        max_length=60,
        blank=False,
        verbose_name=ugettext_lazy('Name'))
    amount = models.DecimalField(
        decimal_places=2,
        max_digits=20,
        blank=False,
        null=False,
        verbose_name=ugettext_lazy('Amount'))
    currency = models.CharField(
        default=hardcode.plan_currency_mxn,
        choices=hardcode.plan_currency,
        max_length=8,
        blank=True,
        verbose_name=ugettext_lazy('Currency'))
    retry_times = models.IntegerField(
        default=3,
        blank=True,
        null=False,
        verbose_name=ugettext_lazy('Number of retries'))
    excerpt = models.CharField(
        max_length=250,
        blank=True,
        verbose_name=ugettext_lazy('Excerpt'))
    description = models.TextField(
        blank=True,
        verbose_name=ugettext_lazy('Description'))
    benefits = JSONField(
        default=dict(),
        blank=True,
        verbose_name=ugettext_lazy('Benefits (as JSON)'))
    status = models.CharField(
        choices=hardcode.plan_status,
        default=hardcode.plan_status_active,
        max_length=9,
        blank=True,
        verbose_name=ugettext_lazy('Status'))
    status_after_retry = models.CharField(
        choices=hardcode.plan_statusafter,
        default=hardcode.plan_statusafter_unpaid,
        max_length=11,
        blank=True,
        verbose_name=ugettext_lazy('Status when retries are exhausted'))
    trial_days = models.IntegerField(
        default=0,
        blank=True,
        null=False,
        verbose_name=ugettext_lazy('Trial days'))
    repeat_every = models.IntegerField(
        default=1,
        blank=True,
        null=False,
        verbose_name=ugettext_lazy('Frecuency Number'))
    repeat_unit = models.CharField(
        choices=hardcode.plan_repeatunit,
        default=hardcode.plan_repeatunit_month,
        max_length=15,
        blank=True,
        verbose_name=ugettext_lazy('Frecuency Unit'))

    @property
    def repeat_verbose(self):
        return ungettext_lazy(
            '%(repeat_unit)s',
            '%(repeat_every)d %(repeat_unit)s',
            'repeat_every') % {
            'repeat_every': self.repeat_every,
            'repeat_unit': self.get_repeat_unit_display(),
        }

    @classmethod
    def get_readonly_fields(self, instance=None):
        if instance:
            return ['openpay_id', 'amount', 'currency', 'retry_times',
                    'status_after_retry', 'repeat_every', 'repeat_unit',
                    'creation_date']
        return ['openpay_id', 'creation_date']

    def op_commit(self):
        if self.openpay_id:
            if not hasattr(self, '_op_'):
                self.op_load()
            if not self.deleted and self._op_:
                self._op_.name = self.name
                self._op_.trial_days = self.trial_days
                self._op_.save()

        else:
            self._op_ = openpay.Plan.create(
                name=self.name,
                amount=str(self.amount),
                currency=self.currency,
                status_after_retry=self.status_after_retry,
                retry_times=self.retry_times,
                repeat_unit=self.repeat_unit,
                trial_days=self.trial_days,
                repeat_every=self.repeat_every)
            self.openpay_id = self._op_.id
        self.op_fill()

    def op_load(self):
        if self.openpay_id:
            try:
                self._op_ = openpay.Plan.retrieve(self.openpay_id)
            except openpay.error.InvalidRequestError as e:
                if e.http_status == 404:
                    self.deleted = True
                    self._op_ = None
                else:
                    raise e
        else:
            raise exceptions.OpenpayObjectDoesNotExist

    def op_fill(self):
        if not hasattr(self, '_op_'):
            self.op_load()
        if self._op_:
            self.op_nones2str(['name'])
            self.name = self._op_.name
            self.amount = Decimal(self._op_.amount)
            self.status_after_retry = self._op_.status_after_retry
            self.retry_times = self._op_.retry_times
            self.repeat_unit = self._op_.repeat_unit
            self.trial_days = self._op_.trial_days
            self.repeat_every = self._op_.repeat_every
            self.creation_date = parse_datetime(
                self._op_.creation_date)

    def __str__(self):
        return self.name


@receiver(pre_save, sender=Plan)
@skippable
def plan_presave(sender, instance=None, **kwargs):
    instance.full_clean()
    instance.op_commit()


@receiver(post_delete, sender=Plan)
@skippable
def plan_postdelete(sender, instance, **kwargs):
    instance.op_dismiss()


class Subscription(AbstractOpenpayBase):
    customer = models.ForeignKey(
        CustomerModel,
        blank=False,
        null=False,
        related_name='subscriptions',
        verbose_name=ugettext_lazy('Customer'))
    card = models.ForeignKey(
        Card,
        blank=False,
        null=False,
        related_name='subscriptions',
        verbose_name=ugettext_lazy('Card'))
    plan = models.ForeignKey(
        Plan,
        blank=False,
        null=False,
        related_name='subscriptions',
        verbose_name=ugettext_lazy('Plan'))
    cancel_at_period_end = models.BooleanField(
        default=False,
        blank=True,
        null=False,
        verbose_name=ugettext_lazy('Cancel at the end of period'))
    latest_charge_date = models.DateField(
        blank=True,
        null=True,
        verbose_name=ugettext_lazy('Previous charge date'))
    charge_date = models.DateField(
        blank=True,
        null=True,
        verbose_name=ugettext_lazy('Next charge date'))
    period_end_date = models.DateField(
        blank=True,
        null=True,
        verbose_name=ugettext_lazy('Period end date'))
    status = models.CharField(
        default=hardcode.subscription_status_trial,
        choices=hardcode.subscription_status,
        max_length=10,
        blank=True,
        verbose_name=ugettext_lazy('Status'))
    current_period_number = models.IntegerField(
        default=0,
        blank=True,
        null=False,
        verbose_name=ugettext_lazy('Trial days'))
    trial_end_date = models.DateField(
        blank=True,
        null=True,
        verbose_name=ugettext_lazy('Trial end date'))

    @classmethod
    def get_readonly_fields(self, instance=None):
        if instance:
            return ['openpay_id', 'customer', 'plan', 'charge_date',
                    'latest_charge_date', 'period_end_date', 'status',
                    'current_period_number', 'creation_date']
        return ['openpay_id', 'charge_date', 'period_end_date', 'status',
                'latest_charge_date', 'current_period_number', 'creation_date']

    def op_commit(self):
        if self.openpay_id:
            if not hasattr(self, '_op_'):
                self.op_load()
            if not self.deleted and self._op_:
                self._op_.trial_end_date = \
                    self.trial_end_date.isoformat()
                self._op_.card = None
                self._op_.card_id = self.card.openpay_id
                self._op_.cancel_at_period_end = \
                    self.cancel_at_period_end
                self._op_.save()

        else:
            if not self.customer or not self.customer.openpay_id:
                raise exceptions.OpenpayNoCustomer
            if not self.card or not self.card.openpay_id:
                raise exceptions.OpenpayNoCard
            self._op_ = openpay.Customer.retrieve(
                self.customer.openpay_id
            ).subscriptions.create(
                plan_id=self.plan.openpay_id,
                trial_end_date=self.trial_end_date.isoformat()
                if self.trial_end_date else None,
                card_id=self.card.openpay_id)
            self.openpay_id = self._op_.id
            if self.cancel_at_period_end:
                self._op_.cancel_at_period_end = \
                    self.cancel_at_period_end
                self._op_.save()
        self.op_fill()

    def op_load(self):
        if not self.customer or not self.customer.openpay_id:
            raise exceptions.OpenpayNoCustomer

        if self.openpay_id:
            try:
                self._op_ = openpay.Customer.retrieve(
                    self.customer.openpay_id
                ).subscriptions.retrieve(self.openpay_id)
            except openpay.error.InvalidRequestError as e:
                if e.http_status == 404:
                    self.deleted = True
                    self._op_ = None
                else:
                    raise e

        else:
            raise exceptions.OpenpayObjectDoesNotExist

    def op_fill(self):
        if not hasattr(self, '_op_'):
            self.op_load()
        if self._op_:
            # self.op_nones2str([])
            self.cancel_at_period_end = \
                self._op_.cancel_at_period_end
            new_charge_date = parse_date(
                self._op_.charge_date)
            self.latest_charge_date = self.charge_date if \
                self.charge_date != new_charge_date else \
                self.latest_charge_date
            self.charge_date = new_charge_date
            self.period_end_date = parse_date(
                self._op_.period_end_date)
            self.status = self._op_.status
            self.current_period_number = self._op_.current_period_number
            self.trial_end_date = parse_date(
                self._op_.trial_end_date)
            self.creation_date = parse_datetime(
                self._op_.creation_date)

    def __str__(self):
        return '{plan} |> {customer}'.format(
            customer=self.customer,
            plan=self.plan)


@receiver(pre_save, sender=Subscription)
@skippable
def subscription_presave(sender, instance=None, **kwargs):
    instance.full_clean()
    if instance.card.customer_id != instance.customer_id:
        raise exceptions.OpenpayNotUserCard
    instance.op_commit()


@receiver(post_delete, sender=Subscription)
@skippable
def subscription_postdelete(sender, instance, **kwargs):
    instance.op_dismiss()


class AbstractTransaction(AbstractOpenpayBase):
    authorization = models.CharField(
        max_length=10,
        blank=True,
        verbose_name=ugettext_lazy('Authorization'))
    transaction_type = models.CharField(
        default=hardcode.transaction_ttype_charge,
        choices=hardcode.transaction_ttype,
        max_length=10,
        blank=True,
        verbose_name=ugettext_lazy('Transaction Type'))
    operation_type = models.CharField(
        default=hardcode.transaction_otype_in,
        choices=hardcode.transaction_otype,
        max_length=4,
        blank=True,
        verbose_name=ugettext_lazy('Operation Type'))
    method = models.CharField(
        default=hardcode.transaction_method_card,
        choices=hardcode.transaction_method,
        max_length=10,
        blank=True,
        verbose_name=ugettext_lazy('Method'))
    order_id = models.CharField(
        max_length=20,
        blank=True,
        verbose_name=ugettext_lazy('Order ID'))
    status = models.CharField(
        default=hardcode.transaction_status_inprogress,
        choices=hardcode.transaction_status,
        max_length=14,
        blank=True,
        verbose_name=ugettext_lazy('Status'))
    amount = models.DecimalField(
        decimal_places=2,
        max_digits=20,
        blank=False,
        null=False,
        verbose_name=ugettext_lazy('Amount'))
    description = models.TextField(
        blank=True,
        verbose_name=ugettext_lazy('Description'))
    error_message = models.TextField(
        blank=True,
        verbose_name=ugettext_lazy('Error Message'))
    customer = models.ForeignKey(
        CustomerModel,
        blank=False,
        null=False,
        verbose_name=ugettext_lazy('Customer'))
    currency = models.CharField(
        default=hardcode.transaction_currency_mxn,
        choices=hardcode.transaction_currency,
        max_length=8,
        blank=True,
        verbose_name=ugettext_lazy('Currency'))
    # TODO: bank_account
    card = models.ForeignKey(
        Card,
        blank=True,
        null=True,
        verbose_name=ugettext_lazy('Card'))
    # TODO: card_points
    operation_date = models.DateTimeField(
        blank=True,
        null=True,
        verbose_name=ugettext_lazy('Operation date'))

    class Meta:
        abstract = True

    @classmethod
    def get_readonly_fields(self, instance=None):
        if instance:
            return ['openpay_id', 'authorization', 'transaction_type',
                    'operation_type', 'method', 'order_id', 'status', 'amount',
                    'description', 'error_message', 'customer', 'currency',
                    'card', 'operation_date', 'creation_date', ]
            # bank_account, card_points
        return ['openpay_id', 'authorization', 'transaction_type',
                'operation_type', 'status', 'error_message', 'operation_date',
                'creation_date', ]


class Charge(AbstractTransaction):
    subscription = models.ForeignKey(
        Subscription,
        blank=True,
        null=True,
        related_name='charges',
        verbose_name=ugettext_lazy('Subscription'))
    conciliated = models.BooleanField(
        default=True,
        blank=True,
        null=False,
        verbose_name=ugettext_lazy('Conciliated'))

    @classmethod
    def get_readonly_fields(self, instance=None):
        if instance:
            return ['conciliated', ] + super().get_readonly_fields(instance)
        return ['conciliated', ] + super().get_readonly_fields(instance)

    def op_capture(self):
        if not self.openpay_id:
            raise exceptions.OpenpayObjectDoesNotExist
        if self.method != hardcode.transaction_method_card:
            raise exceptions.OpenpayNoCard

        if not hasattr(self, '_op_'):
            self.op_load()
        if not self.deleted and self._op_:
            self._op_.capture()

    def op_refund(self, amount=None):
        if not self.openpay_id:
            raise exceptions.OpenpayObjectDoesNotExist
        if self.method != hardcode.transaction_method_card:
            raise exceptions.OpenpayNoCard

        if not hasattr(self, '_op_'):
            self.op_load()
        if not self.deleted and self._op_:
            self._op_.refund(amount) if amount else self._op_.refund()

    def op_commit(self):
        if not self.openpay_id:
            if not self.customer or not self.customer.openpay_id:
                raise exceptions.OpenpayNoCustomer
            if not self.card or not self.card.openpay_id:
                raise exceptions.OpenpayNoCard
            self._op_ = openpay.Customer.retrieve(
                self.customer.openpay_id
            ).charges.create(
                source_id=self.card.openpay_id,
                method=self.method,
                amount=str(self.amount),
                currency=self.currency,
                description=self.description,
                device_session_id=openpay.device_id,
                capture=True)
            self.openpay_id = self._op_.id
        self.op_fill()

    def op_load(self):
        if self.openpay_id:
            if self.customer and self.customer.openpay_id:
                try:
                    self._op_ = openpay.Customer.retrieve(
                        self.customer.openpay_id
                    ).charges.retrieve(self.openpay_id)
                except openpay.error.InvalidRequestError:
                    try:
                        self._op_ = openpay.Charge.retrieve_as_merchant(
                            self.openpay_id)
                    except openpay.error.InvalidRequestError as e:
                        if e.http_status == 404:
                            self.deleted = True
                            self._op_ = None
                        else:
                            raise e

            else:
                try:
                    self._op_ = openpay.Charge.retrieve_as_merchant(
                        self.openpay_id)
                except openpay.error.InvalidRequestError as e:
                    if e.http_status == 404:
                        self.deleted = True
                        self._op_ = None
                    else:
                        raise e

        else:
            raise exceptions.OpenpayObjectDoesNotExist

    def op_fill(self):
        if not hasattr(self, '_op_'):
            self.op_load()
        if self._op_:
            self.op_nones2str([
                'authorization', 'order_id', 'description', 'error_message'])
            self.authorization = self._op_.authorization
            self.operation_type = self._op_.operation_type
            self.transaction_type = self._op_.transaction_type
            self.status = self._op_.status
            self.conciliated = self._op_.conciliated
            self.operation_date = parse_datetime(
                self._op_.operation_date)
            self.description = self._op_.description
            self.error_message = self._op_.error_message
            self.order_id = self._op_.order_id
            self.amount = Decimal(self._op_.amount)
            self.method = self._op_.method
            self.currency = self._op_.currency
            self.creation_date = parse_datetime(
                self._op_.creation_date)
            if hasattr(self._op_, 'subscription_id'):
                self.subscription = Subscription.objects.get(
                    openpay_id=self._op_.subscription_id)
            if hasattr(self._op_, 'customer_id'):
                self.customer = get_customer_model().objects.get(
                    openpay_id=self._op_.customer_id)
            if hasattr(self._op_, 'card'):
                self.card = Card.objects.get(
                    openpay_id=self._op_.card.id)
                if not self.customer and hasattr(
                        self._op_.card, 'customer_id'):
                    self.customer = get_customer_model().objects.get(
                        openpay_id=self._op_.card.customer_id)

    def __str__(self):
        return self.openpay_id


@receiver(pre_save, sender=Charge)
@skippable
def charge_presave(sender, instance=None, **kwargs):
    instance.full_clean()
    if instance.card.customer_id != instance.customer_id:
        raise exceptions.OpenpayNotUserCard
    instance.op_commit()


# This WILL FAIL. And that is the point: to prevent the deletion of charges
# @receiver(pre_delete, sender=Charge)
# def charge_predelete(sender, instance, **kwargs):
#     instance.remove()


class Refund(AbstractTransaction):
    charge = models.OneToOneField(
        Charge,
        blank=False,
        null=False,
        related_name='refund',
        verbose_name=ugettext_lazy('Charge'))
    conciliated = models.BooleanField(
        default=True,
        blank=True,
        null=False,
        verbose_name=ugettext_lazy('Conciliated'))

    @classmethod
    def get_readonly_fields(self, instance=None):
        if instance:
            return ['charge', 'conciliated', ] + \
                super().get_readonly_fields(instance)
        return ['charge', 'conciliated', ] + \
            super().get_readonly_fields(instance)

    def __str__(self):
        return self.openpay_id



from . import ugettext_lazy

address_countrycodes = (  # FIXME: Translate from spanish to english
    ('AF', ugettext_lazy('Afganistán')),
    ('AL', ugettext_lazy('Albania')),
    ('DE', ugettext_lazy('Alemania')),
    ('AD', ugettext_lazy('Andorra')),
    ('AO', ugettext_lazy('Angola')),
    ('AI', ugettext_lazy('Anguila')),
    ('AG', ugettext_lazy('Antigua y Barbuda')),
    ('AN', ugettext_lazy('Antillas Holandesas')),
    ('AQ', ugettext_lazy('Antártida')),
    ('SA', ugettext_lazy('Arabia Saudita')),
    ('DZ', ugettext_lazy('Argelia')),
    ('AR', ugettext_lazy('Argentina')),
    ('AM', ugettext_lazy('Armenia')),
    ('AW', ugettext_lazy('Aruba')),
    ('AU', ugettext_lazy('Australia')),
    ('AT', ugettext_lazy('Austria')),
    ('AZ', ugettext_lazy('Azerbaiyán')),
    ('BS', ugettext_lazy('Bahamas')),
    ('BH', ugettext_lazy('Bahráin')),
    ('BD', ugettext_lazy('Bangladesh')),
    ('BB', ugettext_lazy('Barbados')),
    ('BZ', ugettext_lazy('Belice')),
    ('BJ', ugettext_lazy('Benín')),
    ('BM', ugettext_lazy('Bermudas')),
    ('BY', ugettext_lazy('Bielorrusia')),
    ('BO', ugettext_lazy('Bolivia')),
    ('BQ', ugettext_lazy('Bonaire, San Eustaquio y Saba')),
    ('BA', ugettext_lazy('Bosnia y Hercegovina')),
    ('BW', ugettext_lazy('Botsuana')),
    ('BR', ugettext_lazy('Brasil')),
    ('BN', ugettext_lazy('Brunéi')),
    ('BG', ugettext_lazy('Bulgaria')),
    ('BF', ugettext_lazy('Burkina Faso')),
    ('BI', ugettext_lazy('Burundi')),
    ('BT', ugettext_lazy('Bután')),
    ('BE', ugettext_lazy('Bélgica')),
    ('CV', ugettext_lazy('Cabo Verde')),
    ('KH', ugettext_lazy('Camboya')),
    ('CM', ugettext_lazy('Camerún')),
    ('CA', ugettext_lazy('Canadá')),
    ('TD', ugettext_lazy('Chad')),
    ('CZ', ugettext_lazy('Chequia')),
    ('CL', ugettext_lazy('Chile')),
    ('CN', ugettext_lazy('China')),
    ('CY', ugettext_lazy('Chipre')),
    ('VA', ugettext_lazy('Ciudad del Vaticano')),
    ('CO', ugettext_lazy('Colombia')),
    ('KM', ugettext_lazy('Comores')),
    ('CG', ugettext_lazy('Congo')),
    ('KP', ugettext_lazy('Corea del Norte')),
    ('KR', ugettext_lazy('Corea del Sur')),
    ('CR', ugettext_lazy('Costa Rica')),
    ('CI', ugettext_lazy('Costa de Marfil')),
    ('HR', ugettext_lazy('Croacia')),
    ('CU', ugettext_lazy('Cuba')),
    ('CW', ugettext_lazy('Curaçao')),
    ('DK', ugettext_lazy('Dinamarca')),
    ('DM', ugettext_lazy('Dominica')),
    ('EC', ugettext_lazy('Ecuador')),
    ('EG', ugettext_lazy('Egipto')),
    ('SV', ugettext_lazy('El Salvador')),
    ('AE', ugettext_lazy('Emiratos Árabes Unidos')),
    ('ER', ugettext_lazy('Eritrea')),
    ('SK', ugettext_lazy('Eslovaquia')),
    ('SI', ugettext_lazy('Eslovenia')),
    ('ES', ugettext_lazy('España')),
    ('US', ugettext_lazy('Estados Unidos')),
    ('EE', ugettext_lazy('Estonia')),
    ('ET', ugettext_lazy('Etiopía')),
    ('PH', ugettext_lazy('Filipinas')),
    ('FI', ugettext_lazy('Finlandia')),
    ('FJ', ugettext_lazy('Fiyi')),
    ('FR', ugettext_lazy('Francia')),
    ('GA', ugettext_lazy('Gabón')),
    ('GM', ugettext_lazy('Gambia')),
    ('GE', ugettext_lazy('Georgia')),
    ('GH', ugettext_lazy('Ghana')),
    ('GI', ugettext_lazy('Gibraltar')),
    ('GD', ugettext_lazy('Granada')),
    ('GR', ugettext_lazy('Grecia')),
    ('GL', ugettext_lazy('Groenlandia')),
    ('GP', ugettext_lazy('Guadalupe')),
    ('GU', ugettext_lazy('Guam')),
    ('GT', ugettext_lazy('Guatemala')),
    ('GF', ugettext_lazy('Guayana Francesa')),
    ('GG', ugettext_lazy('Guernsey')),
    ('GN', ugettext_lazy('Guinea')),
    ('GQ', ugettext_lazy('Guinea Ecuatorial')),
    ('GW', ugettext_lazy('Guinea-Bissau')),
    ('GY', ugettext_lazy('Guyana')),
    ('HT', ugettext_lazy('Haití')),
    ('NL', ugettext_lazy('Holanda')),
    ('HN', ugettext_lazy('Honduras')),
    ('HK', ugettext_lazy('Hong Kong')),
    ('HU', ugettext_lazy('Hungría')),
    ('IN', ugettext_lazy('India')),
    ('ID', ugettext_lazy('Indonesia')),
    ('IQ', ugettext_lazy('Iraq')),
    ('IE', ugettext_lazy('Irlanda')),
    ('IR', ugettext_lazy('Irán')),
    ('BV', ugettext_lazy('Isla Bouvet')),
    ('CX', ugettext_lazy('Isla Christmas')),
    ('NF', ugettext_lazy('Isla Norfolk')),
    ('IM', ugettext_lazy('Isla de Man')),
    ('IS', ugettext_lazy('Islandia')),
    ('AX', ugettext_lazy('Islas Aland')),
    ('KY', ugettext_lazy('Islas Caimán')),
    ('CC', ugettext_lazy('Islas Cocos')),
    ('CK', ugettext_lazy('Islas Cook')),
    ('FO', ugettext_lazy('Islas Feroe')),
    ('GS', ugettext_lazy('Islas Georgia del Sur y Sandwich del Sur')),
    ('HM', ugettext_lazy('Islas Heard y McDonald')),
    ('FK', ugettext_lazy('Islas Malvinas')),
    ('MP', ugettext_lazy('Islas Marianas del Norte')),
    ('MH', ugettext_lazy('Islas Marshall')),
    ('PN', ugettext_lazy('Islas Pitcairn')),
    ('SB', ugettext_lazy('Islas Salomón')),
    ('TC', ugettext_lazy('Islas Turcas y Caicos')),
    ('VI', ugettext_lazy('Islas Vírgenes Americanas')),
    ('VG', ugettext_lazy('Islas Vírgenes Británicas')),
    ('UM', ugettext_lazy('Islas menores alejadas de los Estados Unidos')),
    ('IL', ugettext_lazy('Israel')),
    ('IT', ugettext_lazy('Italia')),
    ('JM', ugettext_lazy('Jamaica')),
    ('JP', ugettext_lazy('Japón')),
    ('JE', ugettext_lazy('Jersey')),
    ('JO', ugettext_lazy('Jordania')),
    ('KZ', ugettext_lazy('Kazajstán')),
    ('KE', ugettext_lazy('Kenia')),
    ('KG', ugettext_lazy('Kirguizistán')),
    ('KI', ugettext_lazy('Kiribati')),
    ('KW', ugettext_lazy('Kuwait')),
    ('LA', ugettext_lazy('Laos')),
    ('LS', ugettext_lazy('Lesoto')),
    ('LV', ugettext_lazy('Letonia')),
    ('LR', ugettext_lazy('Liberia')),
    ('LY', ugettext_lazy('Libia')),
    ('LI', ugettext_lazy('Liechtenstein')),
    ('LT', ugettext_lazy('Lituania')),
    ('LU', ugettext_lazy('Luxemburgo')),
    ('LB', ugettext_lazy('Líbano')),
    ('MO', ugettext_lazy('Macao')),
    ('MK', ugettext_lazy('Macedonia')),
    ('MG', ugettext_lazy('Madagascar')),
    ('MY', ugettext_lazy('Malasia')),
    ('MW', ugettext_lazy('Malaui')),
    ('MV', ugettext_lazy('Maldivas')),
    ('MT', ugettext_lazy('Malta')),
    ('ML', ugettext_lazy('Malí')),
    ('MA', ugettext_lazy('Marruecos')),
    ('MQ', ugettext_lazy('Martinica')),
    ('MU', ugettext_lazy('Mauricio')),
    ('MR', ugettext_lazy('Mauritania')),
    ('YT', ugettext_lazy('Mayotte')),
    ('FM', ugettext_lazy('Micronesia')),
    ('MD', ugettext_lazy('Moldavia')),
    ('MN', ugettext_lazy('Mongolia')),
    ('ME', ugettext_lazy('Montenegro')),
    ('MS', ugettext_lazy('Montserrat')),
    ('MZ', ugettext_lazy('Mozambique')),
    ('MM', ugettext_lazy('Myanmar')),
    ('MX', ugettext_lazy('México')),
    ('MC', ugettext_lazy('Mónaco')),
    ('NA', ugettext_lazy('Namibia')),
    ('NR', ugettext_lazy('Nauru')),
    ('NP', ugettext_lazy('Nepal')),
    ('NI', ugettext_lazy('Nicaragua')),
    ('NG', ugettext_lazy('Nigeria')),
    ('NU', ugettext_lazy('Niue')),
    ('NO', ugettext_lazy('Noruega')),
    ('NC', ugettext_lazy('Nueva Caledonia')),
    ('NZ', ugettext_lazy('Nueva Zelanda')),
    ('NE', ugettext_lazy('Níger')),
    ('OM', ugettext_lazy('Omán')),
    ('PW', ugettext_lazy('Palaos')),
    ('PS', ugettext_lazy('Palestina')),
    ('PA', ugettext_lazy('Panamá')),
    ('PG', ugettext_lazy('Papúa New Guinea')),
    ('PK', ugettext_lazy('Paquistán')),
    ('PY', ugettext_lazy('Paraguay')),
    ('PE', ugettext_lazy('Perú')),
    ('PF', ugettext_lazy('Polinesia Francesa')),
    ('PL', ugettext_lazy('Polonia')),
    ('PT', ugettext_lazy('Portugal')),
    ('PR', ugettext_lazy('Puerto Rico')),
    ('QA', ugettext_lazy('Qatar')),
    ('GB', ugettext_lazy('Reino Unido')),
    ('CF', ugettext_lazy('República Centroafricana')),
    ('CD', ugettext_lazy('República Democrática del Congo')),
    ('DO', ugettext_lazy('República Dominicana')),
    ('RE', ugettext_lazy('Reunión')),
    ('RW', ugettext_lazy('Ruanda')),
    ('RO', ugettext_lazy('Rumania')),
    ('RU', ugettext_lazy('Rusia')),
    ('EH', ugettext_lazy('Sahara Occidental')),
    ('WS', ugettext_lazy('Samoa')),
    ('AS', ugettext_lazy('Samoa Americana')),
    ('BL', ugettext_lazy('San Bartolomé')),
    ('KN', ugettext_lazy('San Cristóbal y Nieves')),
    ('SM', ugettext_lazy('San Marino')),
    ('MF', ugettext_lazy('San Martín')),
    ('SX', ugettext_lazy('San Martín (región holandesa)')),
    ('PM', ugettext_lazy('San Pedro y Miquelón')),
    ('VC', ugettext_lazy('San Vicente y las Granadinas')),
    ('SH', ugettext_lazy('Santa Helena')),
    ('LC', ugettext_lazy('Santa Lucia')),
    ('ST', ugettext_lazy('Santo Tomé y Príncipe')),
    ('SN', ugettext_lazy('Senegal')),
    ('RS', ugettext_lazy('Serbia')),
    ('SC', ugettext_lazy('Seychelles')),
    ('SL', ugettext_lazy('Sierra Leona')),
    ('SG', ugettext_lazy('Singapur')),
    ('SY', ugettext_lazy('Siria')),
    ('SO', ugettext_lazy('Somalia')),
    ('LK', ugettext_lazy('Sri Lanka')),
    ('SZ', ugettext_lazy('Suazilandia')),
    ('ZA', ugettext_lazy('Sudáfrica')),
    ('SD', ugettext_lazy('Sudán')),
    ('SS', ugettext_lazy('Sudán del Sur')),
    ('SE', ugettext_lazy('Suecia')),
    ('CH', ugettext_lazy('Suiza')),
    ('SR', ugettext_lazy('Surinam')),
    ('SJ', ugettext_lazy('Svalbard y Jan Mayen')),
    ('TH', ugettext_lazy('Tailandia')),
    ('TW', ugettext_lazy('Taiwán')),
    ('TZ', ugettext_lazy('Tanzania')),
    ('TJ', ugettext_lazy('Tayikistán')),
    ('IO', ugettext_lazy('Territorio Británico del Océano Índico')),
    ('TF', ugettext_lazy('Territorios Franceses del Sur')),
    ('TL', ugettext_lazy('Timor Oriental')),
    ('TG', ugettext_lazy('Togo')),
    ('TK', ugettext_lazy('Tokelau')),
    ('TO', ugettext_lazy('Tonga')),
    ('TT', ugettext_lazy('Trinidad y Tobago')),
    ('TM', ugettext_lazy('Turkmenistán')),
    ('TR', ugettext_lazy('Turquía')),
    ('TV', ugettext_lazy('Tuvalu')),
    ('TN', ugettext_lazy('Túnez')),
    ('UA', ugettext_lazy('Ucrania')),
    ('UG', ugettext_lazy('Uganda')),
    ('UY', ugettext_lazy('Uruguay')),
    ('UZ', ugettext_lazy('Uzbekistán')),
    ('VU', ugettext_lazy('Vanuatu')),
    ('VE', ugettext_lazy('Venezuela')),
    ('VN', ugettext_lazy('Vietnam')),
    ('WF', ugettext_lazy('Wallis y Futuna')),
    ('YE', ugettext_lazy('Yemen')),
    ('DJ', ugettext_lazy('Yibuti')),
    ('ZM', ugettext_lazy('Zambia')),
    ('ZW', ugettext_lazy('Zimbabue')),
)

card_brands_unknown = 'unknown'
card_brands_visa = 'visa'
card_brands_mastercard = 'mastercard'
card_brands_carnet = 'carnet'
card_brands_american = 'american_express'
card_brands_american2 = 'american express'
card_brands = (
    (card_brands_unknown, ugettext_lazy('Unknown')),
    (card_brands_visa, ugettext_lazy('Visa')),
    (card_brands_mastercard, ugettext_lazy('Mastercard')),
    (card_brands_carnet, ugettext_lazy('Carnet')),
    (card_brands_american, ugettext_lazy('American Express')),
    (card_brands_american2, ugettext_lazy('American Express')),
)

plan_currency_mxn = 'MXN'
plan_currency_usd = 'USD'
plan_currency = (
    (plan_currency_mxn, ugettext_lazy('MXN')),
    (plan_currency_usd, ugettext_lazy('USD')),
)
plan_status_active = 'active'
plan_status_deleted = 'deleted'
plan_status = (
    (plan_status_active, ugettext_lazy('Active')),
    (plan_status_deleted, ugettext_lazy('Deleted')),
)
plan_statusafter_unpaid = 'unpaid'
plan_statusafter_cancelled = 'cancelled'
plan_statusafter = (
    (plan_statusafter_unpaid, ugettext_lazy('Unpaid')),
    (plan_statusafter_cancelled, ugettext_lazy('Cancelled')),
)
plan_repeatunit_week = 'week'
plan_repeatunit_month = 'month'
plan_repeatunit_year = 'year'
plan_repeatunit = (
    (plan_repeatunit_week, ugettext_lazy('Week')),
    (plan_repeatunit_month, ugettext_lazy('Month')),
    (plan_repeatunit_year, ugettext_lazy('Year')),
)

subscription_status_trial = 'trial'
subscription_status_active = 'active'
subscription_status_pastdue = 'past_due'
subscription_status_unpaid = 'unpaid'
subscription_status_cancelled = 'cancelled'
subscription_status = (
    (subscription_status_trial, ugettext_lazy('Trial')),
    (subscription_status_active, ugettext_lazy('Active')),
    (subscription_status_pastdue, ugettext_lazy('Past Due')),
    (subscription_status_unpaid, ugettext_lazy('Unpaid')),
    (subscription_status_cancelled, ugettext_lazy('Cancelled')),
)

transaction_ttype_fee = 'fee'
transaction_ttype_charge = 'charge'
transaction_ttype_payout = 'payout'
transaction_ttype_transfer = 'transfer'
transaction_ttype = (
    (transaction_ttype_charge, ugettext_lazy('Charge')),
    # TODO: Activate when models are created
    # (transaction_ttype_fee, ugettext_lazy('Fee')),
    # (transaction_ttype_payout, ugettext_lazy('Payout')),
    # (transaction_ttype_transfer, ugettext_lazy('Transfer')),
)
transaction_otype_in = 'in'
transaction_otype_out = 'out'
transaction_otype = (
    (transaction_otype_in, ugettext_lazy('In')),
    (transaction_otype_out, ugettext_lazy('Out')),
)
transaction_method_card = 'card'
transaction_method_bank = 'bank'
transaction_method_customer = 'customer'
transaction_method_store = 'store'
transaction_method_bankaccount = 'bank_account'
transaction_method = (
    (transaction_method_card, ugettext_lazy('Card')),
    # TODO: Activate when models are created
    # (transaction_method_store, ugettext_lazy('Store')),
    # (transaction_method_bank, ugettext_lazy('Bank Account')),
)
transaction_status_completed = 'completed'
transaction_status_inprogress = 'in_progress'
transaction_status_failed = 'failed'
transaction_status = (
    (transaction_status_completed, ugettext_lazy('Completed')),
    (transaction_status_inprogress, ugettext_lazy('In Progress')),
    (transaction_status_failed, ugettext_lazy('Failed')),
)
transaction_currency_mxn = 'MXN'
transaction_currency_usd = 'USD'
transaction_currency = (
    (transaction_currency_mxn, ugettext_lazy('MXN')),
    (transaction_currency_usd, ugettext_lazy('USD')),
)



from django.db import transaction
from django.shortcuts import render
from django.http import HttpResponse
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_POST

import json

from . import openpay, models, webhooks, decorators


@require_POST
@transaction.atomic
def cardSave(request):
    if request.method == 'POST':
        body = None
        if request.is_ajax():
            body = json.loads(request.body.decode("utf-8"))
        else:
            body = request.POST
        token = body.get('token', False)
        deviceId = body.get('deviceId', False)
        customerId = body.get('customerId', False)
        if token and deviceId and customerId:
            card = models.Card.create_with_token(
                customerId=customerId,
                tokenId=token,
                deviceId=deviceId,
                alias=body.get('alias', ''))
            return HttpResponse()
    return HttpResponse(status=400)


@csrf_exempt
@require_POST
@decorators.basic_auth_required
@transaction.atomic
def webhook(request):
    if request.method == 'POST':
        body = json.loads(request.body.decode("utf-8"))

        callFunctions = {
            'verification': webhooks.verification,
            'charge.refunded': webhooks.chargeRefunded,
            'charge.cancelled': webhooks.chargeCancelled,
            'charge.created': webhooks.chargeCreated,
            'charge.succeeded': webhooks.chargeSucceeded
            # TODO: Add pending webhooks:
            # 'charge.failed',
            # 'charge.rescored.to.decline',
            # 'subscription.charge.failed',
            # 'payout.created',
            # 'payout.succeeded',
            # 'payout.failed',
            # 'transfer.succeeded',
            # 'fee.succeeded',
            # 'fee.refund.succeeded',
            # 'spei.received',
            # 'chargeback.created',
            # 'chargeback.rejected',
            # 'chargeback.accepted',
            # 'order.created',
            # 'order.activated',
            # 'order.payment.received',
            # 'order.completed',
            # 'order.expired',
            # 'order.cancelled',
            # 'order.payment.cancelled',

        }
        option = callFunctions.get(body.get('type', ''), None)

        if option is not None:
            option(body)
        # TODO: Log the webhook
        return HttpResponse()

    return HttpResponse(status=400)



"""
Django settings for django_openpay_repo project.

Generated by 'django-admin startproject' using Django 1.10.1.

For more information on this file, see
https://docs.djangoproject.com/en/1.10/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/1.10/ref/settings/
"""

import os
from django_jinja.builtins import DEFAULT_EXTENSIONS as DJJINJA_DEFAULT

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/1.10/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = os.environ.get('DJANGO_SECRET_KEY')

OPENPAY_PRIVATE_API_KEY = os.environ.get('OPENPAY_PRIVATE_API_KEY')
OPENPAY_PUBLIC_API_KEY = os.environ.get('OPENPAY_PUBLIC_API_KEY')
OPENPAY_MERCHANT_ID = os.environ.get('OPENPAY_MERCHANT_ID')
OPENPAY_DEVICE_ID = os.environ.get('OPENPAY_DEVICE_ID')
OPENPAY_VERIFY_SSL = False
OPENPAY_CUSTOMER_MODEL = 'django_openpay.Customer'

OPENPAY_BASICAUTH_USERS = {
  "daniel": "grvtylabs"
}

# CELERY STUFF
BROKER_URL = 'redis://localhost:6379'
CELERY_RESULT_BACKEND = 'redis://localhost:6379'
CELERY_ACCEPT_CONTENT = ['application/json']
CELERY_TASK_SERIALIZER = 'json'
CELERY_RESULT_SERIALIZER = 'json'
CELERY_TIMEZONE = 'Africa/Nairobi'


# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = []


# Application definition

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'django_openpay',
    'webpack_loader',
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = 'django_openpay_repo.urls'

WSGI_APPLICATION = 'django_openpay_repo.wsgi.application'


# Database
# https://docs.djangoproject.com/en/1.10/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    }
}


# Password validation
# https://docs.djangoproject.com/en/1.10/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]


# Internationalization
# https://docs.djangoproject.com/en/1.10/topics/i18n/
LOCALE_PATHS = (
    os.path.join(BASE_DIR, 'locale'),
)

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.10/howto/static-files/

STATICFILES_DIRS = [
    os.path.join(BASE_DIR, 'django_openpay/static'),
    os.path.join(BASE_DIR, 'testing/static'),
    os.path.join(BASE_DIR, 'webpack'),
]

STATICFILES_FINDERS = (
    'django.contrib.staticfiles.finders.FileSystemFinder',
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',
)

STATIC_URL = '/built/'
STATIC_ROOT = 'built/'

TEMPLATES = [
    {
        'BACKEND': 'django_jinja.backend.Jinja2',
        'DIRS': [
            os.path.join(BASE_DIR, 'django_openpay/templates'),
            os.path.join(BASE_DIR, 'testing/templates'),
        ],
        'APP_DIRS': True,
        'OPTIONS': {
            'match_extension': '.jinja2',
            'app_dirname': 'templates',
            'extensions': DJJINJA_DEFAULT + [
                'django_jinja.builtins.extensions.DjangoFiltersExtension',
                'webpack_loader.contrib.jinja2ext.WebpackExtension',
            ],
            'constants': {
                'OPENPAY_PUBLIC_API_KEY': OPENPAY_PUBLIC_API_KEY,
                'OPENPAY_MERCHANT_ID': OPENPAY_MERCHANT_ID,
            },
            'context_processors': [
                'django.contrib.auth.context_processors.auth',
                'django.template.context_processors.debug',
                'django.template.context_processors.i18n',
                'django.template.context_processors.media',
                'django.template.context_processors.static',
                'django.template.context_processors.tz',
                'django.template.context_processors.request',
                'django.contrib.messages.context_processors.messages',

            ],
        },
    },
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [
        ],
        'APP_DIRS': True,
        'OPTIONS': {
            'debug': DEBUG,
            'context_processors': [
                'django.contrib.auth.context_processors.auth',
                'django.template.context_processors.debug',
                'django.template.context_processors.i18n',
                'django.template.context_processors.media',
                'django.template.context_processors.static',
                'django.template.context_processors.tz',
                'django.template.context_processors.request',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

WEBPACK_LOADER = {
    'DEFAULT': {
        'CACHE': not DEBUG,
        'BUNDLE_DIR_NAME': 'bundles/' if DEBUG else 'production/',
        'STATS_FILE': os.path.join(
                        BASE_DIR, 'webpack/django_openpay_dev.json'
                    ) if DEBUG else os.path.join(
                        BASE_DIR, 'webpack/django_openpay_prod.json'
                    ),
        'POLL_INTERVAL': 0.1,
        'TIMEOUT': None,
    }
}

LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'filters': {
        'require_debug_false': {
            '()': 'django.utils.log.RequireDebugFalse'
        }
    },
    'handlers': {
        'mail_admins': {
            'level': 'ERROR',
            'filters': ['require_debug_false'],
            'class': 'django.utils.log.AdminEmailHandler'
        },
        'console': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
        },
    },
    'loggers': {
        'django.request': {
            'handlers': ['mail_admins'],
            'level': 'ERROR',
            'propagate': True,
        },
    }
}

BROKER_URL = 'redis://localhost:6379'
CELERY_RESULT_BACKEND = 'redis://localhost:6379'
CELERY_ACCEPT_CONTENT = ['application/json']
CELERY_TASK_SERIALIZER = 'json'
CELERY_RESULT_SERIALIZER = 'json'
CELERY_TIMEZONE = 'Africa/Nairobi'
